(* file: rippleCarryAdder.ml
 * author: Bob Muller
 *
 * The positionalAdder code that we covered a few weeks ago can correctly 
 * add positional numerals in any base. It does it's work by applying an 
 * addDigits function, position by position, weaving the carrys from 
 * position to position, right to left. This is all well and good but 
 * the addDigits function is a little problematic: it implements something 
 * simple (digit addition) using operations that are far more complicated: 
 * division and remainder.  Ideally, we would define addition with something 
 * simpler, in fact, something simple enough to realize in computer hardware.
 *
 * We'll restrict out attention to addtion of binary numerals (base 2)
 * and we'll feel free to use 0 and 1 interchangeably with the boolean
 * values false and true. 
 *
 * A truth table for adding bits
 *
 *   a   b   cin    sum   cout
 * +---+---+-----++-----+-----+
 * | 0 | 0 |  0  ||  0  |  0  |
 * +---+---+-----++-----+-----+
 * | 0 | 0 |  1  ||  1  |  0  |
 * +---+---+-----++-----+-----+
 * | 0 | 1 |  0  ||  1  |  0  |
 * +---+---+-----++-----+-----+
 * | 0 | 1 |  1  ||  0  |  1  |
 * +---+---+-----++-----+-----+
 * | 1 | 0 |  0  ||  1  |  0  |
 * +---+---+-----++-----+-----+
 * | 1 | 0 |  1  ||  0  |  1  |
 * +---+---+-----++-----+-----+
 * | 1 | 1 |  0  ||  0  |  1  |
 * +---+---+-----++-----+-----+
 * | 1 | 1 |  1  ||  1  |  1  |
 * +---+---+-----++-----+-----+
 *
 * The output bits on the right are ones when certain combinations
 * of input bits are ones. These situations can be enumerate using
 * logic gates. See, e.g.,
 * 
 * http://tinyurl.com/q9y75l4
 *)

(* xor : bool -> bool -> bool
 *
 * The call (xor a b) computes the exclusive or of a and b.
 *)
let xor a b = (a || b) && (not (a && b))

(* fullAdder : bool -> bool -> bool -> (bool * bool)
 *
 * The call (fullAdder a b carryIn) adds the bits a, b and carryIn
 * producing the sum and the carry out.
 *)
let fullAdder a b cin =
  let c = xor a b in
  let sum = xor c cin in
  let carryOut = (c && cin) || (a && b)
  in
  (sum, carryOut)

(* rippleCarryAdder : bool list -> bool list -> bool -> (bool list * bool)
 *
 * The call (rippleCarryAdder ms ns carryIn) implements a ripple carry
 * adder. It is essentially identical to the positionAdd function.
 *)
let rippleCarryAdder ms ns =
  let rec repeat ms ns carryIn acc =
    match (ms, ns) with
    | ([], []) -> (acc, carryIn)
    | (m::ms, n::ns) -> let (sum, carryOut) = fullAdder m n carryIn
			in
			repeat ms ns carryOut (sum::acc)
  in
  repeat (List.rev ms) (List.rev ns) false []
